/*
 * Copyright 2003 Fruhwirth Clemens 
 */

#include "aes.h"
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/crypto.h>
#include <asm/byteorder.h>
#include <linux/types.h>
#include <linux/crypto.h>

struct twofish_ctx {
	cipherInstance ci;
	keyInstance ki;
};

typedef struct twofish_ctx twofish_ctx_t;

/* Needed for twofish2.c */
int useAsm = 7;

unsigned long Here(unsigned long x)
{
	return 0;
}
/* end */

static int twofish_setkey(void *ctx_, const u8 *key,unsigned int keylen, u32 *flags)
{
	char *hexbuf;
	twofish_ctx_t *ctx = ctx_;
	int i;
	
	hexbuf = kmalloc(keylen*2+1,GFP_KERNEL);	
	cipherInit(&(ctx->ci),MODE_ECB,NULL);
	for(i=0; i<keylen; i++)
		sprintf(hexbuf+(i*2),"%02x",key[i]);	
	makeKey(&(ctx->ki),DIR_ENCRYPT,keylen*8,hexbuf);
	kfree(hexbuf);
	return 0;
}

/* Twofish's assembler implementation only supports in-place encryption
   for some reason. It fails to manage to copy the memory on it's own.
   I don't intent to find that bug, so just do the copying */

static void twofish_decrypt(void *ctx1, u8 *dst, const u8 *src)
{
	twofish_ctx_t *ctx = ctx1;
	if(dst != src) 
		memcpy(dst,src,BLOCK_SIZE/8);
	blockDecrypt(&(ctx->ci),&(ctx->ki),dst,BLOCK_SIZE/8,dst);
}

static void twofish_encrypt(void *ctx1, u8 *dst, const u8 *src)
{
	twofish_ctx_t *ctx = ctx1;
	if(dst != src) 
		memcpy(dst,src,BLOCK_SIZE/8);
	blockEncrypt(&(ctx->ci),&(ctx->ki),dst,BLOCK_SIZE/8,dst);
}
	
static struct crypto_alg alg = {
	.cra_name	= 	"twofish",
	.cra_flags	= 	CRYPTO_ALG_TYPE_CIPHER,
	.cra_blocksize	= 	BLOCK_SIZE/8,
	.cra_ctxsize	= 	sizeof(twofish_ctx_t),
	.cra_module	= 	THIS_MODULE,
	.cra_u		= 	{ 
		.cipher = {
			.cia_min_keysize	= MIN_KEY_BITS/8,
			.cia_max_keysize	= MAX_KEY_BITS/8,
			.cia_setkey		= twofish_setkey,
			.cia_encrypt		= twofish_encrypt,
			.cia_decrypt		= twofish_decrypt,
		}
	},
	.cra_list	=	LIST_HEAD_INIT(alg.cra_list)
};

static int __init init(void)
{
	struct cpuinfo_x86 *c = cpu_data;
	if (c->x86 < 6) 
	{	printk("twofish-i586 loaded: using Pentium code\n");
		useAsm = 7;
	}
	else
	{	printk("twofish-i586 loaded: using PentiumPro code\n");
		useAsm = 15;
	}
	return crypto_register_alg(&alg);
}

static void __exit fini(void)
{
	crypto_unregister_alg(&alg);
}

module_init(init);
module_exit(fini);

MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("Twofish Cipher (i586/i686 Implementation)");
MODULE_AUTHOR("Fruhwirth Clemens");
